/*
*
* mod_gfx.c - The main implementation of mod_gfx
* Copyright (C) 2009 Michael Spiegle (mike@nauticaltech.com)
*	
* This file is part of mod_gfx.
* 
* mod_gfx is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* mod_gfx is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*
*/

//Please don't judge me by my code.  I'm a n00b developer.
#include <mod_gfx.h>
#include <gd.h>
#include <string.h>

#include <httpd.h>
#include <http_log.h>
#include <http_protocol.h>
#include <http_config.h>
#include <http_core.h>
#include <ap_config.h>
#include <ap_regex.h>

#include <apr_strings.h>
#include <apr_signal.h>

static int
gfx_main_handler(request_rec* r) {
	//ignore request if it isn't for us
	if (!r->handler || strcmp(r->handler, "gfx"))
		return DECLINED;
	
	//we only allow GET for now (maybe ever?)
	if (r->method_number != M_GET)
		return gfx_custom_response(r, "mod_gfx: That method is not allowed", HTTP_METHOD_NOT_ALLOWED);

	//disable pesky SIGPIPE.  not sure if this is the right place for this
	//apr_signal(SIGPIPE, SIG_IGN);

	//get server conf
	gfx_server_config_t* sconf = ap_get_module_config(r->server->module_config, &gfx_module);

	//make ourselves a local working copy of the uri
	char* request_uri = apr_pstrdup(r->pool, r->uri);

	//check if URI contains profile
	char* pos = NULL;
	if (NULL == (pos = strchr(request_uri + 1, '/')))
		return gfx_custom_response(r, "mod_gfx: The profile was not specified", HTTP_BAD_REQUEST);
	
	//get profile from URI
	*pos = '\0';
	char* image_profile = apr_pstrcat(r->pool, request_uri + 1, NULL);

	//get path
	char* full_image_path = apr_psprintf(r->pool, "/%s", pos + 1);
	
	//match the full_image_path against our origins to find the hostname
	apr_hash_index_t* iter;
	char found = 0;
	gfx_origin_config_t* origin;
	for (iter = apr_hash_first(r->pool, sconf->origins); iter != NULL; iter = apr_hash_next(iter)) {
		apr_hash_this(iter, NULL, NULL, (void*)&origin);
		//this is to catch origins that don't have a regex specified
		//it would be more efficient to check all configs prior to starting
		//and ensure they have regex's attached to them
		if (origin->regex == NULL)
			continue;
		if (ap_regexec(origin->regex, full_image_path, 0, NULL, AP_REG_EXTENDED) == 0) {
			found = 1;
			break;
		}
	}

	//check to see if an origin was matched or not
	if (found == 0)
		return gfx_custom_response(r, "mod_gfx: The URI is not correct", HTTP_BAD_REQUEST);

	//to support GfxSkipPath, we will skip as many /'s as you want
	if (origin->skippath > 0) {
		int y;
		char* temp;
		for (y = 0; y < origin->skippath; y++) {
			if (NULL == (temp = strchr(full_image_path + 1, '/')))
				return gfx_custom_response(r, "mod_gfx: The URI is not correct", HTTP_INTERNAL_SERVER_ERROR);
			full_image_path = temp;
		}
	}
	char* image_path = apr_pstrdup(r->pool, full_image_path);


	//now that we have an origin, select a host from it
	//todo: make sure all origins have at least one host
	//we'll just pick the first host from the list for now
	const char* origin_hostname = ((const char**)origin->hosts->elts)[0];

	apr_sockaddr_t* addr;
	apr_socket_t* sock;
	apr_status_t result;

	//Generate sockaddr (i kinda like this)
	//todo - I think this returns failure when dns can't be looked up
	if (apr_sockaddr_info_get(&addr, origin_hostname, APR_INET, 80, 0, r->pool) != APR_SUCCESS) {
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "Failed to generate sockaddr");
		return HTTP_INTERNAL_SERVER_ERROR;
	}
	
	//Create socket, or return error
	if (apr_socket_create(&sock, addr->family, SOCK_STREAM,
	                      APR_PROTO_TCP, r->pool) != APR_SUCCESS) {
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "Failed to create socket");
		return HTTP_INTERNAL_SERVER_ERROR;
	}

	//Set some options on the socket
	if (apr_socket_opt_set(sock, APR_TCP_NODELAY, 1) != APR_SUCCESS) {
		apr_socket_close(sock);
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "Failed to set TCP_NODELAY");
		return HTTP_INTERNAL_SERVER_ERROR;
	}
	if (apr_socket_opt_set(sock, APR_TCP_NODELAY, 1) != APR_SUCCESS) {
		apr_socket_close(sock);
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "Failed to set TCP_NODELAY");
		return HTTP_INTERNAL_SERVER_ERROR;
	}

	//Socket timeout
	apr_socket_timeout_set(sock, CON_TIMEOUT);

	//Make connection
	if ((result = apr_socket_connect(sock, addr)) != APR_SUCCESS) {
		apr_socket_close(sock);
		char error[1024];
		apr_strerror(result, error, sizeof(error));
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "Failed to connect to host: %s", error);
		return HTTP_INTERNAL_SERVER_ERROR;
	}

	//Create and send request
	char req[MAX_LEN];
	snprintf(req, MAX_LEN, "%s %s %s\n%s %s\n\n", "GET", 
	         image_path,
			 "HTTP/1.0",
			 "Host:",
			 origin_hostname);
	apr_size_t bw = 0;
	bw = strlen(req);
	if (apr_socket_send(sock, req, &bw) != APR_SUCCESS) {
		apr_socket_close(sock);
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "Failed to send data");
		return HTTP_INTERNAL_SERVER_ERROR;
	}

	//now we need to parse the http response, get the content-length, and the actual data. mao
	http_state_t state = HTTP_PROTO_BEGIN;
	int x;
	char t[MAX_LEN]; //temp storage
	int tp = 0;      //temp storage ptr
	char v[MAX_LEN]; //temp storage for header values
	int vp = 0;      //temp storage ptr

	//allocate storage for image, and register w/ pool for cleanup
	char* b = (char*)malloc(BUF_SIZE);
	apr_pool_cleanup_register(r->pool, b, gfx_pool_cleanup, apr_pool_cleanup_null);
	apr_size_t bp = 0; //body ptr
	int bs = BUF_SIZE; //size of b

	//lets try using an APR table
	apr_table_t* headers = apr_table_make(r->pool, 10);
	
	char resp[BUF_SIZE];
	apr_size_t br = BUF_SIZE;
	while (apr_socket_recv(sock, resp, &br) == APR_SUCCESS) {
		if (br == -1) //this probably shouldn't be happening with apr
			ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "apr_socket_recv says bytesread = -1");

		if (br == 0) //we're done here
			break;
		
		//start switch parser
		for (x = 0; x < br; x++) {

			//we skip \r unless we're parsing content
			//this might suck for windows, but we'll fix that later
			if (resp[x] == '\r' && state != HTTP_PROTO_BODY)
				continue;

			switch(state) {
				case HTTP_PROTO_BEGIN:
					if (resp[x] == '/') {
						t[tp] = '\0';
						apr_table_set(headers, "PROTOCOL", t); //this better be HTTP....
						tp = 0;
						state++;
					} else {
						t[tp] = resp[x];
						tp++;
					}
					break;

				case HTTP_PROTO_VERSION:
					if (resp[x] == ' ') {
						t[tp] = '\0';
						apr_table_set(headers, "VERSION", t);
						tp = 0;
						state++;
					} else {
						t[tp] = resp[x];
						tp++;
					}
					break;

				case HTTP_PROTO_STATUS:
					if (resp[x] == ' ') {
						t[tp] = '\0';
						apr_table_set(headers, "STATUS", t);
						tp = 0;
						state++;
					} else {
						t[tp] = resp[x];
						tp++;
					}
					break;

				case HTTP_PROTO_MSG:
					if (resp[x] == '\n') {
						t[tp] = '\0';
						apr_table_set(headers, "MESSAGE", t);
						tp = 0;
						state++;
					} else {
						t[tp] = resp[x];
						tp++;
					}
					break;

				case HTTP_PROTO_KEY:
					if (resp[x] == ':') {
						t[tp] = '\0';
						tp = 0;
						state++;
					} else {
						t[tp] = resp[x];
						tp++;
					}
					break;

				case HTTP_PROTO_VALUE:
					if (resp[x] == '\n') {
						v[vp] = '\0';
						vp = 0;
						apr_table_set(headers, t, v);
						state++;
					} else {
						if (vp > 1 || resp[x] != ' ') { //avoid starting whitespace
							v[vp] = resp[x];
							vp++;
						}
					}
					break;

				case HTTP_PROTO_HFIN:
					if (resp[x] == '\n') {
						state++;
					} else {
						x--;
						state = HTTP_PROTO_KEY;
					}
					break;

				case HTTP_PROTO_BODY:
					if (bp < bs) {
						b[bp] = resp[x];
						bp++;
					} else {
						//todo: find a better way to handle this
						apr_pool_cleanup_kill(r->pool, b, gfx_pool_cleanup);
						bs += BUF_SIZE;
						b = (char*)realloc(b, bs);
						apr_pool_cleanup_register(r->pool, b, gfx_pool_cleanup, apr_pool_cleanup_null);
						x--;
					}
					break;

				default:
					//this should never happen
					apr_socket_close(sock);
					ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "bad state encountered");
					return HTTP_INTERNAL_SERVER_ERROR;
					break;
			}
		}
	}

	apr_socket_close(sock);

	//Lets make sure we got back a 200 before continuing...
	apr_int64_t status = apr_atoi64(apr_table_get(headers, "STATUS"));
	switch (status) {
		case 200:
			break;

		case 403:
			return gfx_custom_response(r, "mod_gfx: The origin returned error 403 forbidden", HTTP_FORBIDDEN);

		case 404:
			return gfx_custom_response(r, "mod_gfx: The origin returned error 404 not found", HTTP_NOT_FOUND);

		default:
			return gfx_custom_response(r, "mod_gfx: The origin returned an unknown error", HTTP_INTERNAL_SERVER_ERROR);
	}

	//Create GD image from src with on-the-fly type detection
	gdImagePtr srcImgPtr = NULL;
	image_type_t original_image_type;
	if (NULL == (srcImgPtr = gfx_gd_from_blob(b, bp, &original_image_type)))
		return gfx_custom_response(r, "mod_gfx: Source image type is unsupported", HTTP_INTERNAL_SERVER_ERROR);

	//grab the profile object
	gfx_profile_config_t* profile;
	if (NULL == (profile = apr_hash_get(sconf->profiles, image_profile, APR_HASH_KEY_STRING)))
		return gfx_custom_response(r, "mod_gfx: The profile does not exist", HTTP_BAD_REQUEST);

	//Perform the necessary operations
	gdImagePtr dstImgPtr = NULL;
	if (NULL == (dstImgPtr = gfx_process_image(srcImgPtr, profile)))
		return gfx_custom_response(r, "mod_gfx: Error processing image", HTTP_INTERNAL_SERVER_ERROR);

	//output JPG image to memory buffer and set content type
	int imgBufs = 0;
	void* imgBuf = NULL;
	if (profile->output_format == IMAGE_TYPE_SRC)
		profile->output_format = original_image_type;
	switch (profile->output_format) {
		case IMAGE_TYPE_JPG:
			ap_set_content_type(r, "image/jpeg");
			imgBuf = gdImageJpegPtr(dstImgPtr, &imgBufs, profile->qual);
			break;
		case IMAGE_TYPE_PNG:
			ap_set_content_type(r, "image/png");
			imgBuf = gdImagePngPtr(dstImgPtr, &imgBufs);
			break;
		case IMAGE_TYPE_GIF:
			ap_set_content_type(r, "image/gif");
			imgBuf = gdImageGifPtr(dstImgPtr, &imgBufs);
			break;
		case IMAGE_TYPE_SRC:
			//we don't need this... it's handled right above the switch
			break;
		default:
			break;
	}

	//try to detect strange errors
	if (imgBuf == NULL || imgBufs == 0)
		return gfx_custom_response(r, "mod_gfx: Error after processing image", HTTP_INTERNAL_SERVER_ERROR);

	//send image to client - content type should already be set
	ap_rwrite(imgBuf, imgBufs, r);

	//cleanup
	gdFree(imgBuf);

	//free the src image, but only free the
	//dst if it really is a separate image
	gdImageDestroy(srcImgPtr);
	if (srcImgPtr != dstImgPtr)
		gdImageDestroy(dstImgPtr);

	return OK;
}

gdImagePtr
gfx_gd_from_blob(void* b, apr_size_t bp, image_type_t* original_image_type) {
	switch (gfx_image_type(b)) {
		case IMAGE_TYPE_JPG:
			if (original_image_type != NULL)
				*original_image_type = IMAGE_TYPE_JPG;
			return gdImageCreateFromJpegPtr(bp, b);
			break;
	
		case IMAGE_TYPE_GIF:
			if (original_image_type != NULL)
				*original_image_type = IMAGE_TYPE_GIF;
			return gdImageCreateFromGifPtr(bp, b);
			break;
	
		case IMAGE_TYPE_PNG:
			if (original_image_type != NULL)
				*original_image_type = IMAGE_TYPE_PNG;
			return gdImageCreateFromPngPtr(bp, b);
			break;
	
		default:
			return NULL;
			break;
	}
	return NULL;
}

unsigned int
gfx_custom_response(request_rec* r, char* message, unsigned int rc) {
	ap_custom_response(r, rc, message);
	return rc;
}

gdImagePtr
gfx_process_image(gdImagePtr src, gfx_profile_config_t* profile) {
	//first, we need to obtain the size of the original image
	int srcX = gdImageSX(src);
	int srcY = gdImageSY(src);
	int dstX = 0;
	int dstY = 0;
	gdImagePtr dst = NULL;

	//now, we need to figure out the operation and do the right thing
	switch(profile->action) {
		case IMAGE_ACTION_RESIZE:
			//don't enlarge images
			if ((srcX <= profile->rsx) && (srcY <= profile->rsy))
				return src;
			//calculate destination dimensions while preserving AR
			if (profile->rsx >= profile->rsy) {
				dstX = profile->rsx;
				dstY = srcY * ((float)profile->rsx / (float)srcX);
			} else {
				dstY = profile->rsy;
				dstX = srcX * ((float)profile->rsy / (float)srcY);
			}
			if (dstX == 0 || dstY == 0 || srcX == 0 || srcY == 0)
				return NULL;
			dst = gdImageCreateTrueColor(dstX, dstY);
			gdImageCopyResized(dst, src, 0, 0, 0, 0, dstX, dstY, srcX, srcY);
			return dst;
			break;
		case IMAGE_ACTION_RESAMPLE:
			//don't enlarge images
			if ((srcX <= profile->rsx) && (srcY <= profile->rsy))
				return src;
			//calculate destination dimensions while preserving AR
			if (profile->rsx >= profile->rsy) {
				dstX = profile->rsx;
				dstY = srcY * ((float)profile->rsx / (float)srcX);
			} else {
				dstY = profile->rsy;
				dstX = srcX * ((float)profile->rsy / (float)srcY);
			}
			if (dstX == 0 || dstY == 0 || srcX == 0 || srcY == 0)
				return NULL;
			dst = gdImageCreateTrueColor(dstX, dstY);
			gdImageCopyResampled(dst, src, 0, 0, 0, 0, dstX, dstY, srcX, srcY);
			return dst;
			break;
		case IMAGE_ACTION_CROP:
			//stub
			break;
		case IMAGE_ACTION_WATERMARK:
			//We need to make sure that the requested location
			//of the watermark is within the bounds of the image
			if ((profile->locx > srcX) || (profile->locy > srcY)) {
				profile->locx = 0;
				profile->locy = 0;
			}
			//we'll default to lower-right for now
			//until we make a config option for it
			int ww = gdImageSX(profile->watermark);
			int wh = gdImageSY(profile->watermark);
			if (profile->mergepct == 100)
				gdImageCopy(src, profile->watermark, srcX - profile->locx - ww, srcY - profile->locy - wh, 0, 0, ww, wh);
			else
				gdImageCopyMerge(src, profile->watermark, srcX - profile->locx - ww, srcY - profile->locy - wh, 0, 0, ww, wh, profile->mergepct);
			return src;
			break;
		case IMAGE_ACTION_NOOP:
			return src;
			break;
	}
	
	return NULL;
}


static int
gfx_image_type(void* ptr) {
	if (ptr == NULL)
		return -1;
	else if (memcmp(ptr, IMAGE_SIG_JPG, IMAGE_SIG_SIZE_JPG) == 0)
		return IMAGE_TYPE_JPG;
	else if (memcmp(ptr, IMAGE_SIG_GIF, IMAGE_SIG_SIZE_GIF) == 0)
		return IMAGE_TYPE_GIF;
	else if (memcmp(ptr, IMAGE_SIG_PNG, IMAGE_SIG_SIZE_PNG) == 0)
		return IMAGE_TYPE_PNG;
	
	return -1;
}

static apr_status_t
gfx_pool_cleanup(void* ptr) {
	free(ptr);
	return APR_SUCCESS;
}

static apr_status_t
gfx_pool_cleanup_gd(void* ptr) {
	gdImageDestroy((gdImagePtr)ptr);
	return APR_SUCCESS;
}

int
gfx_status_handler(request_rec* r) {
	ap_set_content_type(r, "text/html");
	ap_rputs("Status Page\n", r);
	return 0;
}

static const char*
gfx_config_profile(cmd_parms* cmd, void* p, const char* arg) {
	const char* endp = ap_strrchr_c(arg, '>');
	const char* args;
	char* profile_name;
	ap_conf_vector_t* gfx_conf;
	
	ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "Entering gfx_config_profile callback");

	//create a new profile for each <GfxProfile> encountered
	gfx_conf = ap_create_per_dir_config(cmd->pool);

	const char* err = ap_check_cmd_context(cmd, NOT_IN_LIMIT | NOT_IN_DIRECTORY);

	if (err != NULL)
		return err;

	//get the profile name (argument to <GfxProfile>)
	//do some checks for good measure
	args = apr_pstrndup(cmd->pool, arg, endp - arg);
	if (!args[0])
		return apr_psprintf(cmd->pool, "%s %s", cmd->cmd->name, "> directive requires profile name");

	if (endp == NULL)
		return apr_psprintf(cmd->pool, "%s %s", cmd->cmd->name, "> directive missing closing '>'");

	profile_name = ap_getword_conf(cmd->pool, &args);
	if (!profile_name)
		return apr_psprintf(cmd->pool, "%s %s", cmd->cmd->name, "> error reading profile name");

	//i abusing this by sending <GfxProfile> so the container creator knows what type to give me back
	gfx_profile_config_t* profile = ap_set_config_vectors(cmd->server, gfx_conf, "<GfxProfile>", &gfx_module, cmd->pool);

	//setup profile
	profile->name = apr_pstrdup(cmd->pool, profile_name);
	
	//not sure yet - i hope it iterates over the gfx parms in the <GfxProfile>
	err = ap_walk_config(cmd->directive->first_child, cmd, gfx_conf);
	if (err != NULL)
		return err;

	//add profile to server profiles hashmap
	gfx_server_config_t* sconf = ap_get_module_config(cmd->server->module_config, &gfx_module);
	apr_hash_set(sconf->profiles, profile->name, APR_HASH_KEY_STRING, profile);

	return NULL;
}

static const char*
gfx_config_origin(cmd_parms* cmd, void* p, const char* arg) {
	const char* endp = ap_strrchr_c(arg, '>');
	const char* args;
	char* origin_name;
	ap_conf_vector_t* gfx_conf;
	
	ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "Entering gfx_config_origin callback");

	//create a new origin for each <GfxOrigin> encountered
	gfx_conf = ap_create_per_dir_config(cmd->pool);

	const char* err = ap_check_cmd_context(cmd, NOT_IN_LIMIT | NOT_IN_DIRECTORY);

	if (err != NULL)
		return err;

	//get the origin name (argument to <GfxOrigin>)
	//do some checks for good measure
	args = apr_pstrndup(cmd->pool, arg, endp - arg);
	if (!args[0])
		return apr_psprintf(cmd->pool, "%s %s", cmd->cmd->name, "> directive requires origin name");

	if (endp == NULL)
		return apr_psprintf(cmd->pool, "%s %s", cmd->cmd->name, "> directive missing closing '>'");

	origin_name = ap_getword_conf(cmd->pool, &args);
	if (!origin_name)
		return apr_psprintf(cmd->pool, "%s %s", cmd->cmd->name, "> error reading origin name");

	//apache won't invoke the directives inside of a container until we
	//call ap_walk_config().  unfortunately, i have no idea how apache's
	//config stuff works.  as a result, we'll let apache do its thing
	//then copy the resultant origin into our "real" origin struct.
	//first, lets detect if an origin exists already and set a flag.
	//yes, this is VERY dirty.  please, someone have pity on me here.
	gfx_server_config_t* sconf = ap_get_module_config(cmd->server->module_config, &gfx_module);
	gfx_origin_config_t* realorigin = apr_hash_get(sconf->origins, origin_name, APR_HASH_KEY_STRING);
	char origin_exists = 0;

	//if we already have an origin, just set a flag and we'll deal with it later
	if (realorigin != NULL) {
		origin_exists = 1;
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "Origin already exists for [%s]", realorigin->name);
	}

	//i think this associates the generic config type with my origin config
	gfx_origin_config_t* fakeorigin;
	fakeorigin = ap_set_config_vectors(cmd->server, gfx_conf, "<GfxOrigin>", &gfx_module, cmd->pool);
	
	//setup origin
	fakeorigin->name = apr_pstrdup(cmd->pool, origin_name);
	
	//not sure yet - i hope it iterates over the gfx parms in the <GfxOrigin>
	err = ap_walk_config(cmd->directive->first_child, cmd, gfx_conf);
	if (err != NULL)
		return err;

	//if we already had an origin created,
	//we need to copy the fake origin to it
	//and add the appropriate one to the hash
	if (origin_exists == 1) {
		realorigin->hosts = fakeorigin->hosts;
		realorigin->skippath = fakeorigin->skippath;
		//realorigin->regex = fakeorigin->regex;
		apr_hash_set(sconf->origins, realorigin->name, APR_HASH_KEY_STRING, realorigin);
	} else {
		apr_hash_set(sconf->origins, fakeorigin->name, APR_HASH_KEY_STRING, fakeorigin);
	}

	return NULL;
}

/* No more multi actions for now
static const char*
gfx_profile_actions(cmd_parms* cmd, void* p, const char* arg) {
	gfx_profile_config_t* profile = p;
	if (profile == NULL)
		return apr_psprintf(cmd->pool, "%s %s", cmd->cmd->name, " error getting profile");
	*(const char**)apr_array_push(profile->actions) = apr_pstrdup(cmd->pool, arg);

	ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "Created GfxAction [%s] for profile [%s]", arg, profile->name);

	return NULL;
}
*/

static const char*
gfx_profile_action(cmd_parms* cmd, void* p, const char* arg) {
	gfx_profile_config_t* profile = p;
	if (profile == NULL)
		return apr_psprintf(cmd->pool, "%s %s", cmd->cmd->name, " error getting profile");
	
	if (arg == NULL) {
		return apr_psprintf(cmd->pool, "%s %s", cmd->cmd->name, ": Error determining action");
	} else if(strcmp(arg, "resize") == 0) {
		profile->action = IMAGE_ACTION_RESIZE;
	} else if(strcmp(arg, "resample") == 0) {
		profile->action = IMAGE_ACTION_RESAMPLE;
	} else if(strcmp(arg, "crop") == 0) {
		profile->action = IMAGE_ACTION_CROP;
	} else if(strcmp(arg, "watermark") == 0) {
		profile->action = IMAGE_ACTION_WATERMARK;
	} else if(strcmp(arg, "noop") == 0) {
		profile->action = IMAGE_ACTION_NOOP;
	} else {
		return apr_psprintf(cmd->pool, "%s %s", cmd->cmd->name, " Invalid action specified");
	}

	ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "Created GfxAction [%s] for profile [%s]", arg, profile->name);

	return NULL;
}

static const char*
gfx_profile_rsx(cmd_parms* cmd, void* p, const char* arg) {
	gfx_profile_config_t* profile = p;
	profile->rsx = apr_atoi64(arg);

	return NULL;
}

static const char*
gfx_profile_rsy(cmd_parms* cmd, void* p, const char* arg) {
	gfx_profile_config_t* profile = p;
	profile->rsy = apr_atoi64(arg);

	return NULL;
}

static const char*
gfx_profile_locx(cmd_parms* cmd, void* p, const char* arg) {
	gfx_profile_config_t* profile = p;
	profile->locx = apr_atoi64(arg);

	return NULL;
}

static const char*
gfx_profile_locy(cmd_parms* cmd, void* p, const char* arg) {
	gfx_profile_config_t* profile = p;
	profile->locy = apr_atoi64(arg);

	return NULL;
}

static const char*
gfx_profile_qual(cmd_parms* cmd, void* p, const char* arg) {
	gfx_profile_config_t* profile = p;
	profile->qual = apr_atoi64(arg);

	return NULL;
}

static const char*
gfx_profile_mergepct(cmd_parms* parms, void* p, const char* arg) {
	gfx_profile_config_t* profile = p;
	apr_int64_t size = apr_atoi64(arg);
	if (size < 1 || size > 100)
		return apr_psprintf(parms->pool, "%s %s", parms->cmd->name, ": Merge pct is out of range (0-100)");

	profile->mergepct = size;

	return NULL;
}

static const char*
gfx_profile_watermark(cmd_parms* parms, void* p, const char* arg) {
	gfx_profile_config_t* profile = p;
	
	apr_finfo_t stat;
	apr_file_t* file;
	if (apr_file_open(&file, arg, APR_READ, APR_OS_DEFAULT, parms->pool) != APR_SUCCESS)
		return apr_psprintf(parms->pool, "%s %s [%s]", parms->cmd->name, ": problem opening watermark file", arg);

	if (apr_file_info_get(&stat, APR_FINFO_SIZE, file) != APR_SUCCESS)
		return apr_psprintf(parms->pool, "%s %s [%s]", parms->cmd->name, ": problem reading watermark file", arg);

	if (stat.size < 1)
		return apr_psprintf(parms->pool, "%s %s [%s]", parms->cmd->name, ": Watermark file less than 1 byte!", arg);

	void* temp = apr_pcalloc(parms->temp_pool, stat.size);
	apr_size_t br;
	if (apr_file_read_full(file, temp, stat.size, &br) != APR_SUCCESS)
		return apr_psprintf(parms->pool, "%s %s [%s]", parms->cmd->name, ": Error copying watermark to profile", arg);

	if (br < stat.size)
		return apr_psprintf(parms->pool, "%s %s [%s]", parms->cmd->name, ": Entire watermark couldn't be read", arg);

	if (NULL == (profile->watermark = gfx_gd_from_blob(temp, br, NULL)))
		return apr_psprintf(parms->pool, "%s %s [%s]", parms->cmd->name, ": Watermark file not supported", arg);

	apr_pool_cleanup_register(parms->pool, (void*)(profile->watermark), gfx_pool_cleanup_gd, apr_pool_cleanup_null);

	return NULL;
}

static const char*
gfx_profile_outputformat(cmd_parms* cmd, void* p, const char* arg) {
	gfx_profile_config_t* profile = p;

	if (arg == NULL) {
		return apr_psprintf(cmd->pool, "%s %s", cmd->cmd->name, ": valid image types are JPG, PNG, GIF, SRC");
	} else if (strcmp(arg, "JPG") == 0) {
		profile->output_format = IMAGE_TYPE_JPG;
	} else if (strcmp(arg, "PNG") == 0) {
		profile->output_format = IMAGE_TYPE_PNG;
	} else if (strcmp(arg, "GIF") == 0) {
		profile->output_format = IMAGE_TYPE_GIF;
	} else if (strcmp(arg, "SRC") == 0) {
		profile->output_format = IMAGE_TYPE_SRC;
	} else {
		return apr_psprintf(cmd->pool, "%s %s", cmd->cmd->name, ": valid image types are JPG, PNG, GIF, SRC");
	}

	return NULL;
}

static const char*
gfx_config_originmatch(cmd_parms* cmd, void* p, const char* arg1, const char* arg2) {
	gfx_server_config_t* sconf = ap_get_module_config(cmd->server->module_config, &gfx_module);
	gfx_origin_config_t* origin = apr_hash_get(sconf->origins, arg1, APR_HASH_KEY_STRING);

	ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "Creating GfxOriginMatch for [%s]", arg1);

	//we can't really control the order of GfxOriginMatch directives
	//and the order that the origin containers are specified in configs.
	//to deal with this, we need to check first and possibly create an origin.
	if (origin == NULL) {
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "origin[%s] doesn't exist, creating", arg1);
		origin = (gfx_origin_config_t*)apr_pcalloc(cmd->pool, sizeof(gfx_origin_config_t));
		origin->name = apr_pstrdup(cmd->pool, arg1);
		origin->hosts = apr_array_make(cmd->pool, 16, sizeof(const char*));
		apr_hash_set(sconf->origins, origin->name, APR_HASH_KEY_STRING, origin);
	}

	origin->regex = NULL;
	origin->regex = ap_pregcomp(cmd->pool, arg2, AP_REG_EXTENDED);
	if (origin->regex == NULL)
		return apr_psprintf(cmd->pool, "Your regular expression for [%s] failed", arg1);

	return NULL;
}

static const char*
gfx_config_originhost(cmd_parms* parms, void* p, const char* arg) {
	gfx_origin_config_t* origin = p;
	if (origin == NULL)
		return apr_psprintf(parms->pool, "%s %s", parms->cmd->name, " error getting origin");

	*(const char**)apr_array_push(origin->hosts) = apr_pstrdup(parms->pool, arg);

	ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "Created GfxOriginHost [%s] for origin [%s]", arg, origin->name);

	return NULL;
}

static const char*
gfx_config_skippath(cmd_parms* parms, void* p, const char* arg) {
	gfx_origin_config_t* origin = p;
	if (origin == NULL)
		return apr_psprintf(parms->pool, "Error with GfxSkipPath directive");

	origin->skippath = apr_atoi64(arg);
	if (origin->skippath < 0 || origin->skippath > 100)
		return apr_psprintf(parms->pool, "Error with GfxSkipPath directive");

	ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "%s->skippath = [%d]", origin->name, origin->skippath);

	return NULL;
}

static void*
gfx_create_server_config(apr_pool_t* pool, server_rec* server) {
	gfx_server_config_t* sconf;
	sconf = (gfx_server_config_t*)apr_pcalloc(pool, sizeof(gfx_server_config_t));

	sconf->profiles = apr_hash_make(pool);
	sconf->origins = apr_hash_make(pool);

	ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "Created server config");

	return (void*)sconf;
}

static void*
gfx_create_container_config(apr_pool_t* pool, char* arg) {
	//I'm not really sure why apache calls this function at startup
	//I don't think I need it, so I'm just going to ignore it...
	if (arg == NULL)
		return (void*)NULL;

	//Create a GfxProfile struct on-command and pass it back to the caller
	if (strcmp(arg, "<GfxProfile>") == 0) {
		gfx_profile_config_t* profile;
		profile = (gfx_profile_config_t*)apr_pcalloc(pool, sizeof(gfx_profile_config_t));
		profile->name = NULL;
		profile->output_format = IMAGE_TYPE_SRC;
		profile->action = IMAGE_ACTION_NOOP;
		//no more multi-actions for now
		//apr_array_make(pool, 8, sizeof(const char*));
		profile->rsx = 0;
		profile->rsy = 0;
		profile->locx = 0;
		profile->locy = 0;
		profile->mergepct = 100;
		profile->watermark = NULL;
		//qual WAS set to 0, but i think 50 is a more sensible default
		profile->qual = 50;
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "Created <GfxProfile> container");
		return profile;
	}

	//create a GfxOrigin struct on-command and pass it back to the caller
	if (strcmp(arg, "<GfxOrigin>") == 0) {
		gfx_origin_config_t* origin;
		origin = (gfx_origin_config_t*)apr_pcalloc(pool, sizeof(gfx_origin_config_t));
		origin->name = NULL;
		origin->regex = NULL;
		origin->skippath = 0;
		origin->hosts = apr_array_make(pool, 16, sizeof(const char*));
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "Created <GfxOrigin> container");
		return origin;
	}

	//default case.  shouldn't ever happen
	ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "Error creating container (%s)", arg);
	return (void*)NULL;
}

static void
gfx_register_hooks(apr_pool_t* pool) {
	ap_hook_handler(gfx_main_handler, NULL, NULL, APR_HOOK_MIDDLE);
}

static const
command_rec gfx_cmds[] = {
	AP_INIT_RAW_ARGS("<GfxProfile", gfx_config_profile, NULL, RSRC_CONF,
                     "Container to create profiles"),
	AP_INIT_NO_ARGS("</GfxProfile>", NULL, NULL, RSRC_CONF, "Place Holder"),
	AP_INIT_RAW_ARGS("<GfxOrigin", gfx_config_origin, NULL, RSRC_CONF,
	                 "Container to create origin hosts"),
	AP_INIT_NO_ARGS("</GfxOrigin>", NULL, NULL, RSRC_CONF, "Place Holder"),
	//disable multiple actions for now
	//AP_INIT_ITERATE("GfxActions", gfx_profile_actions, NULL, RSRC_CONF,
	//                "Defines a list of actions to perform on the image"),
	AP_INIT_TAKE1("GfxAction", gfx_profile_action, NULL, RSRC_CONF,
	              "Defines a list of actions to perform on the image"),
	AP_INIT_TAKE1("GfxRsX", gfx_profile_rsx, NULL, RSRC_CONF,
	              "Specifies a maximum width to resize to"),
	AP_INIT_TAKE1("GfxRsY", gfx_profile_rsy, NULL, RSRC_CONF,
	              "Specifies a maximum height to resize to"),
	AP_INIT_TAKE1("GfxLocX", gfx_profile_locx, NULL, RSRC_CONF,
	              "Specifies position for watermark"),
	AP_INIT_TAKE1("GfxLocY", gfx_profile_locy, NULL, RSRC_CONF,
	              "Specifies position for watermark"),
	AP_INIT_TAKE1("GfxWatermark", gfx_profile_watermark, NULL, RSRC_CONF,
                  "Specifies filename for watermark"),
	AP_INIT_TAKE1("GfxQual", gfx_profile_qual, NULL, RSRC_CONF,
	              "Specifies the quality (if supported by output format)"),
	AP_INIT_TAKE1("GfxMergePct", gfx_profile_mergepct, NULL, RSRC_CONF,
	              "Specifies the overlay percent for watermarks"),
	AP_INIT_TAKE1("GfxOutputFormat", gfx_profile_outputformat, NULL, RSRC_CONF,
	              "Specifies the ouput format"),
	AP_INIT_TAKE2("GfxOriginMatch", gfx_config_originmatch, NULL, RSRC_CONF,
	              "A regex to declare origins"),
	AP_INIT_TAKE1("GfxOriginHost", gfx_config_originhost, NULL, RSRC_CONF,
	              "Define a host to origin content from"),
	AP_INIT_TAKE1("GfxSkipPath", gfx_config_skippath, NULL, RSRC_CONF,
	              "Helps you strip off path components"),
	//AP_INIT_TAKE1("GfxTranslate", gfx_config_translate, NULL, RSRC_CONF,
	//              "More flexible way to transate paths"),
	{ NULL }
};

module AP_MODULE_DECLARE_DATA gfx_module = {
	STANDARD20_MODULE_STUFF,
	gfx_create_container_config, //abused container creator
	NULL,                        //i'm too lazy to implement merging
	gfx_create_server_config,    //create server config
	NULL,                        //i'm too lazy to implement merging
	gfx_cmds,                    //command table
	gfx_register_hooks           //hooks
};

	/* header dumper
	ap_set_content_type(r, "text/html; charset=ISO-8859-1");
	ap_rputs("<pre>\n", r);
	ap_rprintf(r, "URI     : %s\n", r->uri);
	ap_rprintf(r, "PROTOCOL: %s\n", apr_table_get(headers, "PROTOCOL"));
	ap_rprintf(r, "VERSION : %s\n", apr_table_get(headers, "VERSION"));
	ap_rprintf(r, "STATUS  : %s\n", apr_table_get(headers, "STATUS"));
	ap_rprintf(r, "MESSAGE : %s\n", apr_table_get(headers, "MESSAGE"));
	ap_rputs("Being Headers...\n", r);
	apr_table_do(gfx_header_printer, (void*)r, headers, NULL);
	ap_rprintf(r, "Content Size: %d\n", cs);
	ap_rputs("</pre>\n", r);
	*/


	/*
	//test dump all config info
	gfx_server_config_t* sconf = ap_get_module_config(r->server->module_config, &gfx_module);
	apr_hash_index_t* iter;

	ap_set_content_type(r, "text/html; charset=ISO-8859-1");
	ap_rputs("<pre>\n", r);

	//dump profiles
	for (iter = apr_hash_first(r->pool, sconf->profiles);
             iter != NULL;
             iter = apr_hash_next(iter)) {

		gfx_profile_config_t* profile;
		apr_hash_this(iter, NULL, NULL, (void*)&profile);
		ap_rprintf(r, "NAME : %s\n", profile->name);
		ap_rprintf(r, "ACTS : %s\n", apr_array_pstrcat(r->pool, profile->actions, ','));
		ap_rprintf(r, "RSX  : %lu\n", profile->rsx);
		ap_rprintf(r, "RXY  : %lu\n", profile->rsy);
		ap_rprintf(r, "Qual : %lu\n", profile->qual);
		ap_rprintf(r, "OutF : %s\n", gfx_output_format(profile->output_format));
		ap_rprintf(r, "\n");
	}

	//dump origins
        for (iter = apr_hash_first(r->pool, sconf->origins);
             iter != NULL;
             iter = apr_hash_next(iter)) {

                gfx_origin_config_t* origin;
                apr_hash_this(iter, NULL, NULL, (void*)&origin);
                ap_rprintf(r, "NAME : %s\n", origin->name);
								ap_rprintf(r, "REGX : %lx\n", (long unsigned)(origin->regex));
                //ap_rprintf(r, "HOSTS: %s\n", apr_array_pstrcat(r->pool, origin->hosts, '\0'));
		int x;
		for (x = 0; x < origin->hosts->nelts; x++) {
			const char* s = ((const char**)origin->hosts->elts)[x];
			ap_rprintf(r, "HOST[%d]: %s\n", x, s);
		}

                ap_rprintf(r, "\n");
        }

	ap_rputs("</pre>\n", r);
	*/

	/*
	ap_rputs("<pre>", r);
	ap_rputs(full_image_path, r);
	ap_rputs("\n", r);
	ap_rputs(image_path, r);
	ap_rputs("\n", r);
	ap_rprintf(r, "skippath = [%d]", origin->skippath);
	ap_rputs("</pre>", r);

	return OK;
	*/
/*
static int
gfx_header_printer(void* data, const char* key, const char* value) {
	request_rec* r = (request_rec*)data;
	ap_rprintf(r, "%s: %s\n", key, value);
	return 1;
}
*/

/*
static char*
gfx_output_format(image_type_t type) {
	switch (type) {
		case IMAGE_TYPE_JPG:
			return "image/jpeg";
			break;
		case IMAGE_TYPE_PNG:
			return "image/png";
			break;
		case IMAGE_TYPE_GIF:
			return "image/gif";
			break;
		case IMAGE_TYPE_SRC:
			return "image/src";
			break;
	}
	return "image/unknown";
}
*/

